
#include <stdio.h>
#include <stdlib.h>

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define LEN	1024

static void	CM_RGB(path,map,R,G,B)
char		*path,*map;
int		*R,*G,*B;
{
	FILE	*file;
	int	pv,r,g,b,c,pv1,pv2;
	char	line[LEN],str[LEN],*s,*brl="bad range list.";

	if (path[0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(path,"r"))==NULL) Error("no colormap file."); 

	for (pv=0; pv<256; pv++) map[pv]=0;

	while (fgets(line,LEN,file)!=NULL) {
	    switch(sscanf(line,"%s %d %d %d",str,&r,&g,&b)) {
		case EOF: goto Break;
		case 2: b=g=r; break;
		case 4: break;
		default: Error("bad color specification.");
	    }
	    if (r<0 || r>=256 ||
		g<0 || g>=256 ||
		b<0 || b>=256) Error("bad R, G and/or B value.");

	    s=str;

#define GETC()	(*(s++))
#define S2N(n)	for (n=0; c>='0' && c<='9'; c=GETC()) \
		    if ((n=n*10+c-'0')>=256) Error(brl)

	    do {
		while ((c=GETC())==',') ;

		if (c!='\0') {
		    if (c=='-') {
			pv1=0;
			if ((c=GETC())==',' || c=='\0')
			    pv2=255;
			else {
			    S2N(pv2);
			    if (c!=',' && c!='\0') Error(brl);
			}
		    }
		    else {
			S2N(pv1);
			if (c==',' || c=='\0')
			    pv2=pv1;
			else if (c=='-')
			    if ((c=GETC())==',' || c=='\0')
				pv2=255;
			    else {
				S2N(pv2);
				if ((c!=',' && c!='\0') || pv2<pv1) Error(brl);
			    }
			else
			    Error(brl);
		    }
		    for (pv=pv1; pv<=pv2; pv++) {
			map[pv]=1; R[pv]=r; G[pv]=g; B[pv]=b;
		    }
		}
	    } while (c==',');
	}
Break:;
	if (file!=stdin) (void)fclose(file);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE		*Fi,*Fo;
	unsigned char	header[13+256*3],*RGB,*rgb;
	int		pvs,pv,R[256],G[256],B[256],c;
	size_t		pvs3;
	double		ratio;
	char		map[256],*dwe="data write error.";

	if (argc!=2 && argc!=4) Error("usage : gif_gcm GIF\n"
				      "        gif_gcm GIF CM_RGB GIF");

	if (argv[1][0]=='-')
	    Fi=stdin;
	else
	    if ((Fi=fopen(argv[1],"rb"))==NULL) Error("GIF file not found.");

	if (fread(header,(size_t)13,(size_t)1,Fi)!=1 ||
	    header[0]!='G' ||
	    header[1]!='I' ||
	    header[2]!='F' ||
	    header[3]!='8' ||
	   (header[4]!='7' && header[4]!='9') ||
	    header[5]!='a') Error("not a GIF file.");

	if (!(header[10]&128)) Error("no global colormap.");

	RGB=header+13; pvs3=(pvs=1<<(header[10]&7)+1)*3;

	if (fread(RGB,pvs3,(size_t)1,Fi)!=1)
	    Error("global colormap read error.");

	ratio=255.0/(double)((1<<((header[10]&112)>>4)+1)-1);

	if (argc==2)
	    for (rgb=RGB, pv=0; pv<pvs; pv++, rgb+=3)
		(void)printf("%d\t%d\t%d\t%d\n",
			     pv,(int)((double)rgb[0]*ratio+0.5),
				(int)((double)rgb[1]*ratio+0.5),
				(int)((double)rgb[2]*ratio+0.5));
	else {
	    if (argv[1][0]=='-' &&
		argv[2][0]=='-') Error("file read confliction.");

	    CM_RGB(argv[2],map,R,G,B);

	    for (rgb=RGB, pv=0; pv<pvs; pv++, rgb+=3)
		if (map[pv]) {
		    rgb[0]=(unsigned char)((double)R[pv]/ratio+0.5);
		    rgb[1]=(unsigned char)((double)G[pv]/ratio+0.5);
		    rgb[2]=(unsigned char)((double)B[pv]/ratio+0.5);
		}

	    if (argv[3][0]=='-')
		Fo=stdout;
	    else
		if ((Fo=fopen(argv[3],"wb"))==NULL)
		    Error("GIF file not stored.");

	    if (fwrite(header,(size_t)13+pvs3,(size_t)1,Fo)!=1) Error(dwe);

	    while ((c=fgetc(Fi))!=EOF) if (fputc(c,Fo)==EOF) Error(dwe);

	    if (Fo!=stdout) (void)fclose(Fo);
	}
	if (Fi!=stdin) (void)fclose(Fi);

	return 0;
}
