
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "rif_rgb.h"
#include "sif.h"
#include "sif_rgb.h"

extern void	Error(char *msg);

typedef struct {
		int	v;
		Cell	r,g,b;
	} Line;

static int	Nx,Ny,z0,gray,Nv,Y;
static Cell	**Z,**fR,**fG,**fB,bR,bG,bB,wR,wG,wB,**mR,**mG,**mB;
static double	unit,V;
static Line	*L;

static void	SlantViewMap(char *SVA,char *path)
{
	double	s,c,v1;
	Line	*l;
	int	x,y,v2,v;
	Cell	r,g,b;

	sincos(atof(SVA)*M_PI/180.0,&s,&c);
	if (s<0.0 || c<0.0) {
	    (void)fprintf(stderr,"%s : bad slant view angle (skipping).\n",SVA);
	    return;
	}
	s*=unit;

	for (l=L, x=0; x<Nx; x++, l++) {
	    l->v=0; l->r=bR; l->g=bG; l->b=bB;
	}
	for (y=0; y<Ny; y++) {
	    v1=V-c*(double)(Y-y);
	    for (l=L, x=0; x<Nx; x++, l++) {
		v2=(int)(v1-s*(double)(Z[y][x]-z0)); r=l->r; g=l->g; b=l->b;
		for (v=l->v; v<v2; v++) {
		    mR[v][x]=r; mG[v][x]=g; mB[v][x]=b;
		}
		l->v=v2; l->r=fR[y][x]; l->g=fG[y][x]; l->b=fB[y][x];
	    }
	}
	for (l=L, x=0; x<Nx; x++, l++) {
	    v=l->v; mR[v][x]=l->r; mG[v][x]=l->g; mB[v][x]=l->b;
	    while (++v<Nv) {
		mR[v][x]=wR; mG[v][x]=wG; mB[v][x]=wB;
	    }
	}
	if (gray)
	    StoreImageFile(path,Nx,Nv,8,mR,NULL);
	else
	    StoreImageFile_RGB(path,Nx,Nv,8,mR,mG,mB,NULL);
}

#define LEN	2048

int	main(int argc,char **argv)
{
	int	x,y,BPS,z,v;
	double	f;
	Cell	r,g,b;
	char	str[LEN],SVA[LEN],path[LEN];

	if (argc!=10 && argc!=12)
	    Error(
	"usage : sv_map value.tif unit SRM.tif bR bG bB wR wG wB {SVA SVM.tif}"
	    );

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&Z,NULL);

	if ((unit=atof(argv[2]))<0.0) Error("bad value unit.");

	ReadImageFile_RGB(argv[3],&x,&y,&BPS,&fR,&fG,&fB,NULL);
	if (Nx!=x || Ny!=y) Error("image size not match.");

	if ((bR=(Cell)atoi(argv[4]))>255 ||
	    (bG=(Cell)atoi(argv[5]))>255 ||
	    (bB=(Cell)atoi(argv[6]))>255 ||
	    (wR=(Cell)atoi(argv[7]))>255 ||
	    (wG=(Cell)atoi(argv[8]))>255 ||
	    (wB=(Cell)atoi(argv[9]))>255) Error("bad R, G and/or B.");

	z=z0=Z[0][0];
	f=255.0/(double)((1<<BPS)-1);
	gray=bR==bG && bG==bB && wR==wG && wG==wB;
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    if (z<Z[y][x]) z=Z[y][x]; else if (z0>Z[y][x]) z0=Z[y][x];

	    r=fR[y][x]=(Cell)(f*(double)fR[y][x]+0.5);
	    g=fG[y][x]=(Cell)(f*(double)fG[y][x]+0.5);
	    b=fB[y][x]=(Cell)(f*(double)fB[y][x]+0.5);

	    gray&=(r==g && g==b);
	}
	Nv=(int)(V=ceil(hypot((double)(Y=Ny-1),unit*(double)(z-z0))))+1;

	if ((L=(Line *)malloc(sizeof(Line)*Nx))==NULL ||
	    (mR=(Cell **)malloc(sizeof(Cell *)*Nv))==NULL ||
	    (mG=(Cell **)malloc(sizeof(Cell *)*Nv))==NULL ||
	    (mB=(Cell **)malloc(sizeof(Cell *)*Nv))==NULL ||
	    (*mR=(Cell *)malloc(sizeof(Cell)*Nv*Nx))==NULL ||
	    (*mG=(Cell *)malloc(sizeof(Cell)*Nv*Nx))==NULL ||
	    (*mB=(Cell *)malloc(sizeof(Cell)*Nv*Nx))==NULL)
	    Error("memory allocation error.");

	for (v=1; v<Nv; v++) {
	    mR[v]=mR[v-1]+Nx; mG[v]=mG[v-1]+Nx; mB[v]=mB[v-1]+Nx;
	}
	if (argc==12)
	    SlantViewMap(argv[10],argv[11]);
	else
	    while (fgets(str,LEN,stdin)!=NULL)
		if (sscanf(str,"%s %s",SVA,path)==2) SlantViewMap(SVA,path);

	free(*mB); free(*mG); free(*mR); free(mB); free(mG); free(mR); free(L);

	for (y=0; y<Ny; y++) {
	    free(fB[y]); free(fG[y]); free(fR[y]); free(Z[y]);
	}
	free(fB); free(fG); free(fR); free(Z);

	return 0;
}
